/*
 * Jalview - A Sequence Alignment Editor and Viewer
 * Copyright (C) 2005 AM Waterhouse, J Procter, G Barton, M Clamp, S Searle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package jalview.bin;

import jalview.gui.*;

import org.apache.log4j.*;

import javax.swing.*;

import java.util.Vector;


/**
 * Main class for Jalview Application
 * <br>
 * <br>start with java -Djava.ext.dirs=$PATH_TO_LIB$ jalview.bin.Jalview
 *
 * @author $author$
 * @version $Revision: 1.29 $
 */
public class Jalview
{

    /**
     * main class for Jalview application
     *
     * @param args open <em>filename</em>
     */
    public static void main(String[] args)
    {
      System.out.println("Java version: "+System.getProperty("java.version"));
      System.out.println(System.getProperty("os.arch")+" "
                         +System.getProperty("os.name")+" "
                         +System.getProperty("os.version"));


      ArgsParser aparser = new ArgsParser(args);
      boolean headless = false;

      if( aparser.contains("help") || aparser.contains("h") )
      {
        System.out.println("Usage: jalview -open [FILE] [OUTPUT_FORMAT] [OUTPUT_FILE]\n\n"
                           +"-nodisplay\tRun Jalview without User Interface.\n"
                           +"-props FILE\tUse the given Jalview properties file instead of users default.\n"
                           +"-groups FILE\tUse the given file to mark groups on the alignment. \nGroups file is in the following tab delimited format\n"
                           +"TEXT<tab>SEQUENCE_ID<tab>SEQUENCE_INDEX<tab>START_RESIDUE<tab>END_RESIDUE<tab>COLOUR\n"
                           +"SequenceID is used in preference to SequenceIndex if both are provided.\n"
                           +"Enter ID_NOT_SPECIFIED for SEQUENCE_ID or -1 for SEQUENCE_INDEX if unknown.\n"
                           +"COLOUR can be hexadecimal RGB or 'red', 'blue' etc.\n\n"
                           +"-fasta FILE\tCreate alignment file FILE in Fasta format.\n"
                           +"-clustal FILE\tCreate alignment file FILE in Clustal format.\n"
                           +"-pfam FILE\tCreate alignment file FILE in PFAM format.\n"
                           +"-msf FILE\tCreate alignment file FILE in MSF format.\n"
                           +"-pileup FILE\tCreate alignment file FILE in Pileup format\n"
                           +"-pir FILE\tCreate alignment file FILE in PIR format.\n"
                           +"-blc FILE\tCreate alignment file FILE in BLC format.\n"
                           +"-jalview FILE\tCreate alignment file FILE in Jalview format.\n"
                           +"-png FILE\tCreate PNG image FILE from alignment.\n"
                           +"-imgMap FILE\tCreate HTML file FILE with image map of PNG image.\n"
                           +"-eps FILE\tCreate EPS file FILE from alignment.");
        System.exit(0);
          }


          if (aparser.contains("nodisplay"))
            System.setProperty("java.awt.headless", "true");

          if (System.getProperty("java.awt.headless") != null
              && System.getProperty("java.awt.headless").equals("true"))
          {
            headless = true;
          }

          try
          {
            Cache.initLogger();
          }
          catch (java.lang.NoClassDefFoundError error)
          {
            error.printStackTrace();
            System.out.println(
                "\nEssential logging libraries not found."
                +"\nUse: java -Djava.ext.dirs=$PATH_TO_LIB$ jalview.bin.Jalview");
            System.exit(0);
          }

        Cache.loadProperties(aparser.getValue("props"));

        Desktop desktop = null;
        if( !headless )
        {
          try
          {
            UIManager.setLookAndFeel(
                   UIManager.getSystemLookAndFeelClassName()
            //        UIManager.getCrossPlatformLookAndFeelClassName()
//"com.sun.java.swing.plaf.gtk.GTKLookAndFeel"
//"javax.swing.plaf.metal.MetalLookAndFeel"
//"com.sun.java.swing.plaf.windows.WindowsLookAndFeel"
//"com.sun.java.swing.plaf.motif.MotifLookAndFeel"

                );
          }
          catch (Exception ex)
          {
          }

          desktop = new Desktop();
          desktop.setVisible(true);
          desktop.discoverer.start();
        }


         String file = null, protocol = null, format = null, groups=null;
         jalview.io.FileLoader fileLoader = new jalview.io.FileLoader();

          file = aparser.getValue("open");

          if (file == null && desktop==null)
          {
            System.out.println("No files to open!");
            System.exit(1);
          }

          if(file!=null)
          {

            if (!file.startsWith("http://"))
            {
              if (! (new java.io.File(file)).exists())
              {
                System.out.println("Can't find " + file);
                System.exit(1);
              }
            }

            protocol = "File";

            if (file.indexOf("http:") > -1)
            {
              protocol = "URL";
            }

            if (file.endsWith(".jar"))
              format = "Jalview";
            else
              format = jalview.io.IdentifyFile.Identify(file, protocol);

            System.out.println("Opening: " + format + " file " + file);

            AlignFrame af = fileLoader.LoadFileWaitTillLoaded(file, protocol, format);

            groups = aparser.getValue("groups");
            if (groups != null)
            {
              af.parseGroupsFile(groups);
            }

            String imageName = "unnamed.png";
            while (aparser.getSize() > 1)
            {
              format = aparser.nextValue();
              file = aparser.nextValue();

              if (format.equalsIgnoreCase("png"))
              {
                af.createPNG(new java.io.File(file));
                imageName = (new java.io.File(file)).getName();
                System.out.println("Creating PNG image: " + file);
                continue;
              }
              else if (format.equalsIgnoreCase("imgMap"))
              {
                af.createImageMap(new java.io.File(file), imageName);
                System.out.println("Creating image map: " + file);
                continue;
              }
              else if (format.equalsIgnoreCase("eps"))
              {
                System.out.println("Creating EPS file: " + file);
                af.createEPS(new java.io.File(file));
                continue;
              }

              if (af.saveAlignment(file, format))
                System.out.println("Written alignment in " + format +
                                   " format to " + file);
              else
                System.out.println("Error writing file " + file + " in " + format +
                                   " format!!");

            }

            while (aparser.getSize() > 0)
            {
              System.out.println("Unknown arg: " + aparser.nextValue());
            }
          }

        // We'll only open the default file if the desktop is visible.
        //////////////////////
          if (
              !headless &&
              jalview.bin.Cache.getDefault("SHOW_STARTUP_FILE", true))
          {

            file = jalview.bin.Cache.getDefault("STARTUP_FILE",
                                                "http://www.jalview.org/examples/exampleFile.jar");
            protocol = "File";

            if (file.indexOf("http:") > -1)
            {
              protocol = "URL";
            }

            if (file.endsWith(".jar"))
            {
              Jalview2XML.LoadJalviewAlign(file);
            }
            else
            {
              format = jalview.io.IdentifyFile.Identify(file, protocol);
              desktop.LoadFile(file, protocol, format);
            }
          }


          if (desktop==null)
            System.exit(0);
    }
}

 class ArgsParser
 {
   Vector vargs = null;
   public ArgsParser(String [] args)
   {
     vargs = new Vector();
     for (int i = 0; i < args.length; i++)
     {
       String arg = args[i].trim();
       if (arg.charAt(0) == '-')
         arg = arg.substring(1);
       vargs.addElement(arg);
     }
   }

   public String getValue(String arg)
   {
     int index = vargs.indexOf(arg);
     String ret = null;
     if (index != -1)
     {
       ret = vargs.elementAt(index + 1).toString();
       vargs.removeElementAt(index);
       vargs.removeElementAt(index);
     }
     return ret;
   }

   public boolean contains(String arg)
   {
     if(vargs.contains(arg))
     {
       vargs.removeElement(arg);
       return true;
     }
     else
       return false;
   }

   public String nextValue()
   {
     return  vargs.remove(0).toString();
   }

   public int getSize()
   {
     return vargs.size();
   }

 }
