/*
 * Jalview - A Sequence Alignment Editor and Viewer
 * Copyright (C) 2005 AM Waterhouse, J Procter, G Barton, M Clamp, S Searle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package jalview.ws;

import java.util.*;

import javax.swing.*;

import ext.vamsas.*;
import jalview.analysis.*;
import jalview.datamodel.*;
import jalview.gui.*;

public class JPredClient
    extends WSClient
{
  ext.vamsas.Jpred server;
  String altitle = "";
  java.util.Hashtable SequenceInfo = null;
  public JPredClient(ext.vamsas.ServiceHandle sh, String title, SequenceI seq) {
    wsInfo = setWebService(sh);
    startJPredClient(title, seq);
  }
  public JPredClient(ext.vamsas.ServiceHandle sh, String title, SequenceI[] msa) {
    wsInfo = setWebService(sh);
    startJPredClient(title, msa);
  }

  public JPredClient(String title, SequenceI[] msf) {
    startJPredClient(title, msf);
  }

  public JPredClient(String title, SequenceI seq) {
    startJPredClient(title, seq);
  }

  private void startJPredClient(String title, SequenceI[] msf)
  {
    if (wsInfo==null)
      wsInfo = setWebService();

    SequenceI seq = msf[0];
    altitle = "JNet prediction on " + seq.getName() +
        " using alignment from " + title;

    wsInfo.setProgressText("Job details for MSA based prediction (" +
                           title + ") on sequence :\n>" + seq.getName() + "\n" +
                           AlignSeq.extractGaps("-. ", seq.getSequence()) +
                           "\n");
    SequenceI aln[] = new SequenceI[msf.length];
    for (int i=0,j=msf.length; i<j;i++)
      aln[i] = new jalview.datamodel.Sequence(msf[i]);
    SequenceInfo = jalview.analysis.SeqsetUtils.uniquify(aln, true);

    if (!locateWebService())
    {
      return;
    }

    JPredThread jthread = new JPredThread(aln);
    jthread.start();
  }

  public void startJPredClient(String title, SequenceI seq)
  {
    if (wsInfo==null)
      wsInfo = setWebService();
    wsInfo.setProgressText("Job details for prediction on sequence :\n>" +
                           seq.getName() + "\n" +
                           AlignSeq.extractGaps("-. ", seq.getSequence()) +
                           "\n");
    altitle = "JNet prediction for sequence " + seq.getName() + " from " +
        title;

    SequenceInfo = jalview.analysis.SeqsetUtils.SeqCharacterHash(seq);

    if (!locateWebService())
    {
      return;
    }

    JPredThread jthread = new JPredThread(seq);
    jthread.start();
  }

  private WebserviceInfo setWebService()
  {
    WebServiceName = "JNetWS";
    WebServiceJobTitle = "JNet secondary structure prediction";
    WebServiceReference =
        "\"Cuff J. A and Barton G.J (1999) Application of enhanced " +
        "multiple sequence alignment profiles to improve protein secondary structure prediction, " +
        "Proteins 40:502-511\".";
    WsURL = "http://www.compbio.dundee.ac.uk/JalviewWS/services/jpred";

    WebserviceInfo wsInfo = new WebserviceInfo(WebServiceJobTitle,
                                               WebServiceReference);

    return wsInfo;
  }

  private boolean locateWebService()
  {
    ext.vamsas.JpredServiceLocator loc = new JpredServiceLocator(); // Default

    try
    {
      this.server = loc.getjpred(new java.net.URL(WsURL)); // JBPNote will be set from properties
      ( (JpredSoapBindingStub)this.server).setTimeout(60000); // one minute stub
    }
    catch (Exception ex)
    {
      JOptionPane.showMessageDialog(Desktop.desktop,
                                    "The Secondary Structure Prediction Service named " +
                                    WebServiceName + " at " + WsURL +
                                    " couldn't be located.",
                                    "Internal Jalview Error",
                                    JOptionPane.WARNING_MESSAGE);
      wsInfo.setProgressText("Serious! " + WebServiceName +
                             " Service location failed\nfor URL :" + WsURL +
                             "\n" +
                             ex.getMessage());
      wsInfo.setStatus(wsInfo.STATE_STOPPED_SERVERERROR);

      return false;
    }

    return true;
  }

  class JPredThread
      extends Thread
  {
    String OutputHeader;
    vamsas.objects.simple.JpredResult result;
    vamsas.objects.simple.Sequence sequence;
    vamsas.objects.simple.Msfalignment msa;
    String jobId;
    boolean jobComplete = false;
    int allowedServerExceptions = 3; // thread dies if too many exceptions.

    JPredThread(SequenceI seq)
    {
      OutputHeader = wsInfo.getProgressText();
      this.sequence = new vamsas.objects.simple.Sequence();
      this.sequence.setId(seq.getName());
      this.sequence.setSeq(AlignSeq.extractGaps("-. ", seq.getSequence()));
    }

    JPredThread(SequenceI[] msf)
    {
      OutputHeader = wsInfo.getProgressText();
      this.sequence = new vamsas.objects.simple.Sequence();
      this.sequence.setId(msf[0].getName());
      this.sequence.setSeq(AlignSeq.extractGaps("-. ",
                                                msf[0].getSequence()));

      jalview.io.PileUpfile mwrite = new jalview.io.PileUpfile();
      this.msa = new vamsas.objects.simple.Msfalignment();
      msa.setMsf(mwrite.print(msf));
    }

    public void run()
    {
      StartJob();

      while (!jobComplete && (allowedServerExceptions > 0))
      {
        try
        {
          if ( (result = server.getresult(jobId)) == null)
          {
            throw (new Exception(
                "Timed out when communicating with server\nTry again later.\n"));
          }
          if (result.getState()==0)
            jalview.bin.Cache.log.debug("Finished "+jobId);
          if (result.isRunning())
          {
            wsInfo.setStatus(WebserviceInfo.STATE_RUNNING);
          }
          if (result.isQueued())
          {
            wsInfo.setStatus(WebserviceInfo.STATE_QUEUING);
          }

          wsInfo.setProgressText(OutputHeader + "\n" +
                                 result.getStatus());

          if (result.isFinished())
          {

            parseResult();
            jobComplete = true;
            jobsRunning--;
          } else {
            // catch exceptions
            if (! (result.isJobFailed() || result.isServerError()))
            {
              try
              {
                Thread.sleep(5000);
              }
              catch (InterruptedException ex1)
              {
              }

              //  System.out.println("I'm alive "+seqid+" "+jobid);
            }
            else
            {
              wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_ERROR);
              jobsRunning--;
              jobComplete = true;
            }
          }
        }
        catch (Exception ex)
        {
          allowedServerExceptions--;
          wsInfo.appendProgressText("\nJPredWS Server exception!\n" +
                                    ex.getMessage());

          try
          {
            if (allowedServerExceptions > 0)
            {
              Thread.sleep(5000);
            }
          }
          catch (InterruptedException ex1)
          {
          }
        }
        catch (OutOfMemoryError er)
        {
          jobComplete = true;
          wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_ERROR);
          JOptionPane.showInternalMessageDialog(Desktop.desktop,
                                                "Out of memory handling result!!"
                                                +
              "\nSee help files for increasing Java Virtual Machine memory."
                                                , "Out of memory",
                                                JOptionPane.WARNING_MESSAGE);
          System.out.println("JPredClient: "+er);
          System.gc();
        }
      }
      if (result!=null)
        if (! (result.isJobFailed() || result.isServerError()))
        {
          wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_OK);
        }
        else
        {
          wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_ERROR);
        }
    }

    void StartJob()
    {
      try
      {
        if (msa != null)
        {
          jobId = server.predictOnMsa(msa);
        }
        else
        {
          jobId = server.predict(sequence);
        }

        if (jobId != null)
        {
          if (jobId.startsWith("Broken"))
          {
            throw new Exception("Submission " + jobId);
          }
          else
          {
            System.out.println(WsURL + " Job Id '" + jobId + "'");
          }
        }
        else
        {
          throw new Exception("Server timed out - try again later\n");
        }
      }
      catch (Exception e)
      {
        if (e.getMessage().indexOf("Exception")>-1) {
          wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_SERVERERROR);
          wsInfo.setProgressText(
              "Failed to submit the prediction. (Just close the window)\n"
              +
              "It is most likely that there is a problem with the server.\n");
          System.err.println(
              "JPredWS Client: Failed to submit the prediction. Quite possibly because of a server error - see below)\n" +
              e.getMessage() + "\n");

          jalview.bin.Cache.log.warn("Server Exception",e);
        } else {
          wsInfo.setStatus(WebserviceInfo.STATE_STOPPED_ERROR);
          // JBPNote - this could be a popup informing the user of the problem.
          wsInfo.setProgressText("Failed to submit the prediction:\n"
                                 +e.getMessage()+
                                 wsInfo.getProgressText());

          jalview.bin.Cache.log.debug("Failed Submission",e);

        }
        allowedServerExceptions = -1;
        jobComplete = true;

      }
    }



    private void addFloatAnnotations(Alignment al, int[] gapmap,
                                     Vector values, String Symname,
                                     String Visname, float min,
                                     float max, int winLength)
    {
      Annotation[] annotations = new Annotation[al.getWidth()];

      for (int j = 0; j < values.size(); j++)
      {
        float value = Float.parseFloat(values.get(j).toString());
        annotations[gapmap[j]] = new Annotation("", value + "", ' ',
                                                value);
      }

      al.addAnnotation(new AlignmentAnnotation(Symname, Visname,
                                               annotations, min, max, winLength));
    }

    void parseResult()
    {
      // OutputHeader = output.getText();
      if (result.isFailed())
      {
        OutputHeader += "Job failed.\n";
      }

      if (result.getStatus() != null)
      {
        OutputHeader += ("\n" + result.getStatus());
      }

      if (result.getPredfile() != null)
      {
        OutputHeader += ("\n" + result.getPredfile());

        // JBPNote The returned files from a webservice could be hidden behind icons in the monitor window that, when clicked, pop up their corresponding data
      }

      if (result.getAligfile() != null)
      {
        OutputHeader += ("\n" + result.getAligfile());
      }

      wsInfo.setProgressText(OutputHeader+"Parsing...");

      try
      {
        jalview.bin.Cache.log.debug("Parsing output from JNet job.");
        // JPredFile prediction = new JPredFile("C:/JalviewX/files/jpred.txt", "File");
        jalview.io.JPredFile prediction = new jalview.io.JPredFile(result.
            getPredfile(),
            "Paste");
        SequenceI[] preds = prediction.getSeqsAsArray();
        jalview.bin.Cache.log.debug("Got prediction profile.");
        Alignment al;
        int FirstSeq; // the position of the query sequence in Alignment al
        boolean noMsa = true; // set if no MSA has been returned by JPred

        if ( (this.msa != null) && (result.getAligfile() != null))
        {
          jalview.bin.Cache.log.debug("Getting associated alignment.");
          // we ignore the returned alignment if we only predicted on a single sequence
          String format = jalview.io.IdentifyFile.Identify(result.getAligfile(),
              "Paste");

          if (jalview.io.FormatAdapter.formats.contains(format))
          {
            al = new Alignment(jalview.io.FormatAdapter.readFile(
                result.getAligfile(), "Paste", format));
            for (int i=0, j=al.getHeight(); i<j; i++) {
              SequenceI sq = al.getSequenceAt(i);
              if (!jalview.analysis.SeqsetUtils.SeqCharacterUnhash(
                  sq, (Hashtable) SequenceInfo.get(sq.getName())))
              {
                throw (new Exception(
                    "Couldn't recover sequence properties for JNet "
                    +((i==0) ? "Query sequence" : "alignment sequence ("+i+")")));
              }
            }
            noMsa = false;
            FirstSeq = 0;
          }
          else
          {
            throw (new Exception(
                "Unknown format 'format' for file : \n" +
                result.getAligfile()));
          }
        }
        else
        {
          al = new Alignment(preds);
          FirstSeq = prediction.getQuerySeqPosition();
          if (!jalview.analysis.SeqsetUtils.SeqCharacterUnhash(
              al.getSequenceAt(FirstSeq), SequenceInfo))
          {
            throw (new Exception(
                "Couldn't recover sequence properties for JNet Query sequence!"));
          }
        }


        AlignmentAnnotation annot;
        Annotation[] annotations = null;
        int i = 0;
        int width = preds[0].getSequence().length();

        int[] gapmap = al.getSequenceAt(FirstSeq).gapMap();

        if (gapmap.length != width)
        {
          throw (new Exception(
              "Jnet Client Error\nNumber of residues in supposed query sequence :\n" +
              al.getSequenceAt(FirstSeq).getName() + "\n" +
              al.getSequenceAt(FirstSeq).getSequence() +
              "\nDiffer from number of prediction sites in \n" +
              result.getPredfile() + "\n"));
        }

        // JBPNote Should also rename the query sequence sometime...
        i = 0;

        while (i < preds.length)
        {
          String id = preds[i].getName().toUpperCase();

          if (id.startsWith("LUPAS") || id.startsWith("JNET") ||
              id.startsWith("JPRED"))
          {
            annotations = new Annotation[al.getWidth()];

            if (id.equals("JNETPRED") || id.equals("JNETPSSM") ||
                id.equals("JNETFREQ") || id.equals("JNETHMM") ||
                id.equals("JNETALIGN") || id.equals("JPRED"))
            {
              for (int j = 0; j < width; j++)
              {
                annotations[gapmap[j]] = new Annotation("", "",
                    preds[i].getCharAt(j), 0);
              }
            }
            else if (id.equals("JNETCONF"))
            {
              for (int j = 0; j < width; j++)
              {
                float value = Float.parseFloat(preds[i].getCharAt(
                    j) + "");
                annotations[gapmap[j]] = new Annotation(preds[i].getCharAt(
                    j) + "", "", preds[i].getCharAt(j),
                    value);
              }
            }
            else
            {
              for (int j = 0; j < width; j++)
              {
                annotations[gapmap[j]] = new Annotation(preds[i].getCharAt(
                    j) + "", "", ' ', 0);
              }
            }

            if (id.equals("JNETCONF"))
            {
              annot = new AlignmentAnnotation(preds[i].getName(),
                                              "JNet Output", annotations, 0f,
                                              10f, 1);
            }
            else
            {
              annot = new AlignmentAnnotation(preds[i].getName(),
                                              "JNet Output", annotations);
            }

            al.addAnnotation(annot);

            if (noMsa)
            {
              al.deleteSequence(preds[i]);
            }
          }

          i++;
        }

        Hashtable scores = prediction.getScores();

        /*  addFloatAnnotations(al, gapmap,  (Vector)scores.get("JNETPROPH"),
                              "JnetpropH", "Jnet Helix Propensity", 0f,1f,1);

          addFloatAnnotations(al, gapmap,  (Vector)scores.get("JNETPROPB"),
         "JnetpropB", "Jnet Beta Sheet Propensity", 0f,1f,1);

          addFloatAnnotations(al, gapmap,  (Vector)scores.get("JNETPROPC"),
                              "JnetpropC", "Jnet Coil Propensity", 0f,1f,1);
         */

        wsInfo.setProgressText(OutputHeader);
        jalview.bin.Cache.log.debug("Finished parsing output.");
        AlignFrame af = new AlignFrame(al);

        Desktop.addInternalFrame(af, altitle,
                                 AlignFrame.NEW_WINDOW_WIDTH,
                                 AlignFrame.NEW_WINDOW_HEIGHT);
      }
      catch (Exception ex)
      {
        jalview.bin.Cache.log.warn("Exception whilst parsing JNet style secondary structure prediction.");
        jalview.bin.Cache.log.debug("Exception: ",ex);
      }
    }
  }
}
